/* -------------------------------------------------------------------	*/
/* Copyright (C) 2000 Robert R. Bruner.  All rights reserved.		*/
/* Contact the author at						*/
/*                    rrb@math.wayne.edu				*/
/*                    Dept of Mathematics				*/
/*                    Wayne State University				*/
/*                    Detroit, Michigan  48202				*/
/*                    USA						*/
/*									*/
/* Version 1.6								*/
/* -------------------------------------------------------------------- */


#include <stdio.h>
#include <stdlib.h>

#include <errors.h>
#include <structur.h>
#include <storage.h>
#include <steen.h>
#include <tensor.h>
#include <io.h>
#include <util.h>

/* ---------------- Tensor IO ----------------------------------------- */
/*  Tensors                                                             */
/*  --------                                                            */
/*                                                                      */
/* The linear form of a tensor (as in a file) is:                       */
/*                                                                      */
/*  num_terms                                                           */
/*  s1 gen1 t1 seq1 opdeg1                                                */
/*  E1                                                                  */
/*  s2 gen2 t2 seq2 opdeg2                                                */
/*  E2                                                                  */
/*  ...                                                                 */
/*                                                                      */
/*  That is, its (header,prefix,body) is (1 int, 5 int, 1 element)      */
/*                                                                      */
/* representing the sum for i = 1 to num_terms of                       */
/*           op_i * gen_i (x) Ei        (where (x) = tensor)            */
/*                                                                      */
/* where filt(gen_i) = si,                                              */
/*         deg(op_i) = opdegi                                           */
/*       seqno(op_i) = seqi                                             */
/* deg(gen_i) + degi = ti                                               */
/* -------------------------------------------------------------------- */

/* ---------------------- fread_shape ------------------------------   */

Tensor_shape * fread_shape(filename,s) char *filename; int s;
{
int i,j,fs;
Tensor_shape *shape;
FILE *Shape;
char message[100];

if ((Shape = fopen(filename,"r")) == NULL)
{  sprintf(message,"\nFile not found: Shape\n");
   terminate(message,BAD_FILE);
}

fscanf(Shape,"%d", &fs);
if (fs < s)
{  sprintf(message,"Shape file good only through filt %d.  Need %d",
           fs,s);
   terminate(message,BAD_FILE);
}

if (s < 0) s = fs;
shape = (Tensor_shape *) malloc(sizeof(Tensor_shape));
shape->s = s++;
shape->numgen = (int *) malloc( s * sizeof(int *));
shape->deg = (int **) malloc( s * sizeof(int *));

for (i=0; i < s; i++) fscanf(Shape,"%d",&shape->numgen[i]);

for (; i <= fs; i++) fscanf(Shape,"%*d");

for (i=0; i < s; i++)
{  shape->deg[i] =
       (int *) malloc (shape->numgen[i] * sizeof(int *));

   for (j=0; j< shape->numgen[i]; j++)
      fscanf(Shape,"%d",&shape->deg[i][j]);
}
fclose(Shape);
return(shape);
}

/* ---------------------- read_shape ------------------------------   */

Tensor_shape * read_shape(s) int s;
{ return(fread_shape("Shape",s));
}

/* ---------------------- read_Tensor ------------------------------   */

Tensor * read_Tensor(fp)  FILE *fp;
{
int  i,deg;
unsigned gen,length,num;
Tensor *x;

if (fscanf(fp,"%u",&num) == EOF) return(NULL);
          /* EOF here just means there are no more elements to read so   */
          /*  we return NULL to indicate that.                           */

x = get_Tensor(num);
for (i=0; i<num; i++)
 {  x->item[i] = get_TTERM();
    if (fscanf(fp,"%d %u %d %u %d",
                 &x->item[i]->filt,
                 &x->item[i]->gen,
                 &x->item[i]->deg,
                 &x->item[i]->seq,
                 &x->item[i]->opdeg) == EOF)
       terminate("Premature EOF in read_Tensor",BAD_READ);
          /* EOF here is in mid element so is an error                   */
    x->item[i]->RHS = read_Element(fp);
 }

return(x);
}

/* ---------------------- skip_Tensor ------------------------------   */

void skip_Tensor(fp)  FILE *fp;
{
int  i,deg;
unsigned gen,length,num;

if (fscanf(fp,"%u",&num) == EOF) return;
          /* EOF here just means there are no more elements to read so   */
          /*  we return NULL to indicate that.                           */

for (i=0; i<num; i++)
 {
    if (fscanf(fp,"%*d %*u %*d %*u %*d") == EOF)
       terminate("Premature EOF in skip_Tensor",BAD_READ);
          /* EOF here is in mid element so is an error                   */
    skip_Element(fp);
 }

return;
}

/* ---------------------- write_Tensor ------------------------------   */

write_Tensor(fp,e,type) FILE *fp;
                        Tensor *e;
                        char type;
{ int i;

if (e == NULL) { fprintf(fp,"\n0\n"); return(0); }

fprintf(fp,"\n%d\n",e->len);
for (i=0; i < e->len; i++)
  { fprintf(fp,"\n%d %u %d %u %d\n",
                e->item[i]->filt,
                e->item[i]->gen,
                e->item[i]->deg,
                e->item[i]->seq,
                e->item[i]->opdeg);
    write_Element(fp,e->item[i]->RHS,type);
  }
return(0);
}

/* ---------------------- read_exp_Tensor ------------------------------   */

Element ***** read_exp_Tensor(fp,px_shape)  FILE *fp;
                                   Tensor_shape *px_shape;
{
   /*      Create an expanded tensor of shape px_shape and read the         */
   /*          next tensor in fp into it, and return it to the caller.      */
int  i,s,t,opdeg;
unsigned g,seq,num;
char message[100];
Element *e;
Element ***** x = NULL;

if (fscanf(fp,"%u",&num) == EOF) return(NULL);
          /* EOF here just means there are no more elements to read so   */
          /*  we return NULL to indicate that.                           */

for (i=0; i<num; i++)
 { if (fscanf(fp,"%d %u %d %u %d",&s,&g,&t,&seq,&opdeg) == EOF)
        terminate("Premature EOF in read_exp_Tensor",BAD_READ);
          /* EOF here is in mid element so is an error                   */
   if (t != opdeg + px_shape->deg[s][g])
      { sprintf(message,
         "\nread_exp_Tensor found inconsistency:s,g,t,seq,opdeg,gendeg = %d %u %d %u %d %d\n",
                s,g,t,seq,opdeg,px_shape->deg[s][g]);
        terminate(message,BAD_READ);
      }
   x = create_node(x,px_shape,s,g,opdeg,seq);
   if ( x[s][g][opdeg][seq] != NULL)
     { sprintf(message,
         "\nDuplicate term s,g,t,seq,opdeg = %d %u %d %u %d",
         s,g,t,seq,opdeg);
       terminate(message,BAD_READ);
     }
   x[s][g][opdeg][seq] = read_Element(fp);
 }

return(x);
}

/* ---------------------- write_exp_Tensor ------------------------   */

write_exp_Tensor(fp,e,pe_shape,type) FILE *fp;
                                Element *****e;
                        Tensor_shape *pe_shape;
                                    char type;
{ int s,g,opdeg,seq,num;
Internal_Op Op;

if (e == NULL) { fprintf(fp,"\n0\n"); return(0); }

/* Determine the number of nonzero entries                             */

for (s=0,num=0; s<= pe_shape->s; s++)
  if (e[s] != NULL)

    for (g=0; g < pe_shape->numgen[s]; g++)
      if (e[s][g] != NULL)
        for (opdeg=0; opdeg <= pe_shape->t - pe_shape->deg[s][g]; opdeg++)
          if (e[s][g][opdeg] != NULL)
            for (seq=0; seq < dimA(opdeg); seq++)
              if (e[s][g][opdeg][seq] != NULL) num++;

/* Now print them                                                     */

fprintf(fp,"\n%d\n",num);

for (s=0; s<= pe_shape->s; s++)
  if (e[s] != NULL)
    for (g=0; g < pe_shape->numgen[s]; g++)
      if (e[s][g] != NULL)
        for (opdeg=0; opdeg <= pe_shape->t - pe_shape->deg[s][g]; opdeg++)
          if (e[s][g][opdeg] != NULL)
            for (seq=0; seq < dimA(opdeg); seq++)
              if (e[s][g][opdeg][seq] != NULL)
                { fprintf(fp,"\n%d %d %d %d %d\n",
                          s,g,opdeg + pe_shape->deg[s][g],seq,opdeg);
                  write_Element(fp,e[s][g][opdeg][seq],type);
                }

return(0);
}

/* ----------------------------------------------------------- */
/* create_node creates a path to the s,g,t,seq entry of r, and */
/*   returns a pointer to the (possibly new or modified) r     */
/* ----------------------------------------------------------- */

Element ***** create_node(r,pr_shape,s,g,opdeg,seq)
                  Element ***** r;
                  Tensor_shape *pr_shape;
                  int s,g,opdeg,seq;
{
int i;

if (r == NULL)
  { r = (Element *****) get_pointers(1+pr_shape->s);
    for (i=0; i <= pr_shape->s; i++) r[i] = NULL;
  }

if (r[s] == NULL)
  { r[s] = (Element ****) get_pointers(pr_shape->numgen[s]);
    for (i=0; i < pr_shape->numgen[s]; i++) r[s][i] = NULL;
  }

if (r[s][g] == NULL)
  { r[s][g] = (Element ***) get_pointers(1+pr_shape->t - pr_shape->deg[s][g]);
    for (i=0; i <= pr_shape->t - pr_shape->deg[s][g]; i++)
         r[s][g][i] = NULL;
  }

if (r[s][g][opdeg] == NULL)
  { r[s][g][opdeg] = (Element **) get_pointers(dimA(opdeg));
    for (i=0; i < dimA(opdeg); i++) r[s][g][opdeg][i] = NULL;
  }

return(r);
}

/* ----------------------------------------------------------- */
/* Tact applies Op to the tensor e using the coproduct of Op   */
/*  and adds the result to the expanded tensor r whose shape   */
/*  is pointed to by pr_shape.                                 */
/* Tact returns the new expanded tensor, so that if r is NULL  */
/*  when Tact is called, it's new value is passed back to the  */
/*  caller.  In all other cases the value is redundant, in     */
/*  that the change is internal to the object pointed to by r  */
/* Tact will replace any terminal node whose element is zero   */
/*  by the NULL pointer.                                       */
/* ----------------------------------------------------------- */

Element *****
    Tact(Op,e,r,pr_shape)
               Internal_Op Op;
               Tensor  *e;
               Tensor_shape *pr_shape;
               Element ***** r;
{

int i,valid;
Internal_Op Op1,Op2;

for (valid = first_Coproduct(Op,Op1,Op2);
       valid;
         valid = next_Coproduct(Op,Op1,Op2))
for (i = 0; i < e->len; i++)
{
                    /* Compute left = Op1 * Opi,  ---------------  */
                    /*    where e->item[i] = opi * geni (x) xi     */
int deg, dOp1 = degree(Op1);
TERM *left;
TTERM *ei = e->item[i];
Internal_Op Lop;

left = get_TERM(dimA(deg = dOp1 + ei->opdeg));
zero_op(left);
left->gen = ei->gen;
left->deg = deg;

first_op(Lop,ei->opdeg);
adv_op(Lop,ei->seq);

mult(Op1,Lop,left);
if (is_not_zero(left))
{
                  /* Compute right = Op2 * xi      --------------- */
Element *right;
int j, rlen, seq;

right = get_Element(ei->RHS->len);
for (j=rlen=0; j < right->len; j++)
{
                 /* Compute jth TERM of right     ----------------- */

TERM *rj,*eij;
Internal_Op Rop;

eij = ei->RHS->item[j];
rj = right->item[j] = get_TERM(dimA(deg = degree(Op2) + eij->deg));
zero_op(rj);
rj->gen = eij->gen;
rj->deg = deg;

for (seq = first_op_in(eij,Rop);
      seq >= 0;
        seq = next_op_in(eij,Rop,seq))
  mult(Op2,Rop,rj);

if (is_not_zero(rj)) rlen++;
}
              /*  right is now computed.  Is it nonzero               */
if (rlen > 0)
{
              /*  Yes, add it to each term of r indexed by a term     */
              /* of left (which is nonzero here)                      */

for (seq = first_seqno_in(left);
      seq >= 0;
        seq = next_seqno_in(left,seq))
    {
      r = create_node(r, pr_shape, ei->filt, ei->gen, dOp1 + ei->opdeg, seq);
      if (r[ei->filt][ei->gen][dOp1 + ei->opdeg][seq] == NULL)
          r[ei->filt][ei->gen][dOp1 + ei->opdeg][seq] = clean_copy(right,rlen);
      else
         add ( &r[ei->filt][ei->gen][dOp1 + ei->opdeg][seq], right);
    }
}

return_whole_Element(right);

}       /* end Compute right                                   */

return_TERM(left);

}       /* end compute left                                      */

return(r);
}

